SystemOrganization addCategory: #UnixMetaStack!

StackInterpreterPrimitives subclass: #MetaStackInterpreter
	instanceVariableNames: ''
	classVariableNames: 'AllButExperimentalBit ExperimentalObjectBit'
	poolDictionaries: ''
	category: 'UnixMetaStack'!

!MetaStackInterpreter class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/5/2013 19:27'!
initialize

	super initialize.
	BytesPerWord ifNil: [BytesPerWord := 4].  "May get called on fileIn, so supply default"	
	WordMask := (1 bitShift: BytesPerWord*8) - 1.
	ExperimentalObjectBit := 1 bitShift: BytesPerWord *8 - 3.  "Next-to-Top bit"
	AllButExperimentalBit  := WordMask - ExperimentalObjectBit.! !

!MetaStackInterpreter class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/6/2013 12:07'!
initializePrimitiveTable

	super initializePrimitiveTable.
	
	PrimitiveTable at: 150 + 1 put: #primitiveGetExperimentalBit.
	PrimitiveTable at: 151 + 1 put: #primitiveSetExperimentalBit.
	PrimitiveTable at: 152 + 1 put: #primitiveTurnOffExperimentalBitOfAllObjects.! !

!MetaStackInterpreter methodsFor: 'meta-protocol-overrides' stamp: 'NickPapoulias 5/6/2013 12:26'!
commonSend
	"Send a message, starting lookup with the receiver's class."
	"Assume: messageSelector and argumentCount have been set, and that 
	the receiver and arguments have been pushed onto the stack,"
	"Note: This method is inlined into the interpreter dispatch loop."
	<sharedCodeNamed: 'commonSend' inCase: #singleExtendedSendBytecode>
	

	(self shouldOverrideBehaviorFor: (self internalStackValue: argumentCount)) ifTrue: [ "rcvr"
		^ self sendMetaSendFor: (self internalStackValue: argumentCount).
	].
	
	(objectMemory splObj: 63) == (self internalStackValue: argumentCount ) ifFalse: ["disable arg meta inside Interpreter"
		0 to: argumentCount - 1 do: [:argIndex | 
			 (self shouldOverrideBehaviorFor: (self internalStackValue: argIndex)) ifTrue: [
					^ self sendMetaArgFor: (self internalStackValue: argIndex)
			]
		]
	].	
	
	super commonSend.
	! !

!MetaStackInterpreter methodsFor: 'experimental-bit-accesors' stamp: 'NickPapoulias 4/30/2013 17:58'!
getExperimentalBitOf: anOop
	self inline: true.
	^((objectMemory baseHeader: anOop) bitAnd: ExperimentalObjectBit) ~= 0 ! !

!MetaStackInterpreter methodsFor: 'experimental-bit-primitives' stamp: 'NickPapoulias 5/5/2013 19:07'!
primitiveGetExperimentalBit
	| rcvr hdr |
	rcvr := self stackTop.
	(self isIntegerObject: rcvr)
		ifTrue:
			[self primitiveFail ]
		ifFalse:
			[ self pop: 1.  "we only pop the receiver"
			self pushBool: (self getExperimentalBitOf: rcvr) ]! !

!MetaStackInterpreter methodsFor: 'experimental-bit-primitives' stamp: 'NickPapoulias 5/5/2013 19:29'!
primitiveSetExperimentalBit
	| boolean rcvr hdr wasImmutable |
	rcvr := self stackValue: 1.
	(self isIntegerObject: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrInappropriate].
	boolean := self booleanValueOf: self stackTop.
	self successful ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	self setExperimentalBitOf: rcvr to: boolean.
	self pop: 1
! !

!MetaStackInterpreter methodsFor: 'experimental-bit-primitives' stamp: 'NickPapoulias 5/5/2013 19:07'!
primitiveTurnOffExperimentalBitOfAllObjects
	| oop |
	oop := self firstAccessibleObject.
	[oop = nil] whileFalse: [
		(self isIntegerObject:  oop)
			ifFalse: [ 
				self setExperimentalBitOf: oop to: false
				].
		oop := self accessibleObjectAfter: oop.
	].
	
! !

!MetaStackInterpreter methodsFor: 'print-debugging' stamp: 'NickPapoulias 4/30/2013 18:01'!
print: s
	
	<api>
	<var: #s type: #'char *'>
	self cCode: 'fputs(s, stdout); fflush(stdout);'! !

!MetaStackInterpreter methodsFor: 'meta-protocol-overrides' stamp: 'NickPapoulias 5/5/2013 17:04'!
pushLiteralVariable: literalIndex

	"place holder for hooking metaRef if needed"

	super pushLiteralVariable: literalIndex.
! !

!MetaStackInterpreter methodsFor: 'meta-protocol-overrides' stamp: 'NickPapoulias 5/5/2013 17:05'!
pushReceiverBytecode

	"place holder for hooking metaRef if needed"

	super pushReceiverBytecode.
! !

!MetaStackInterpreter methodsFor: 'meta-protocol-overrides' stamp: 'NickPapoulias 5/5/2013 17:05'!
pushReceiverVariable: fieldIndex

	|iv object|

	self inline: true.
	
	object := self receiver.
	
	(self shouldOverrideBehaviorFor: object) ifTrue: [
		self sendMetaReadAtFor: object withIndex: (self integerObjectOf: fieldIndex + 1). 
	] ifFalse: [
		super pushReceiverVariable: fieldIndex.
	].
	
	
! !

!MetaStackInterpreter methodsFor: 'meta-protocol-overrides' stamp: 'NickPapoulias 5/5/2013 17:06'!
pushTemporaryVariable: temporaryIndex

	super pushTemporaryVariable: temporaryIndex.

	
! !

!MetaStackInterpreter methodsFor: 'meta-protocol' stamp: 'NickPapoulias 5/5/2013 17:07'!
sendMetaArgFor: object

	"hail mary implementation, rcvr is pushed after args but still works"

	self inline: true.

	self unsetOverrideBehaviorFor: object.
	self externalizeIPandSP.
	self createActualMessageTo: lkupClass.
	self internalPush: object.
	messageSelector := objectMemory splObj: 62. 
	argumentCount := 2.
	self lookupMethodInClass: lkupClass.
	self internalizeIPandSP.
	self internalExecuteNewMethod.
	self fetchNextBytecode.! !

!MetaStackInterpreter methodsFor: 'meta-protocol' stamp: 'NickPapoulias 5/5/2013 17:03'!
sendMetaReadAtFor: object withIndex: index

	self inline: true.

	self unsetOverrideBehaviorFor: object.
	self internalPush: object.
	self internalPush: index.
	messageSelector := objectMemory splObj: 60.
	argumentCount := 1.
	self undoFetchNextBytecode. "hail marry hack"
	self normalSend.! !

!MetaStackInterpreter methodsFor: 'meta-protocol' stamp: 'NickPapoulias 4/30/2013 18:16'!
sendMetaReferenceFor: object 

	self inline: true.
	
	self unsetOverrideBehaviorFor: object.
	self internalPush: object.
	messageSelector := objectMemory splObj: 59.
	argumentCount := 0.
	self normalSend.! !

!MetaStackInterpreter methodsFor: 'meta-protocol' stamp: 'NickPapoulias 5/5/2013 17:02'!
sendMetaSendFor: object

	self inline: true.

	self unsetOverrideBehaviorFor: object.
	self externalizeIPandSP.
	self createActualMessageTo: lkupClass.
	messageSelector := objectMemory splObj: 58. 
	self lookupMethodInClass: lkupClass.
	self internalizeIPandSP.
	self internalExecuteNewMethod.
	self fetchNextBytecode.! !

!MetaStackInterpreter methodsFor: 'meta-protocol' stamp: 'NickPapoulias 5/5/2013 17:03'!
sendMetaWriteAtPutFor: object withIndex: index andValue: value 

	self inline: true.

	self unsetOverrideBehaviorFor: object.
	self internalPush: object.
	self internalPush: index.
	self internalPush: value.
	messageSelector := objectMemory splObj: 61.
	argumentCount := 2.
	self normalSend.! !

!MetaStackInterpreter methodsFor: 'experimental-bit-accesors' stamp: 'NickPapoulias 4/30/2013 17:59'!
setExperimentalBitOf: anOop to: boolean
	| hdr |
	self inline: true.
	"Dont check here if it is integer. Check in the sender of this."
	hdr := objectMemory baseHeader: anOop.
	boolean 
		ifTrue: [ objectMemory baseHeader: anOop put: (hdr bitOr: ExperimentalObjectBit). ]
		ifFalse: [ objectMemory baseHeader: anOop put: (hdr bitAnd: AllButExperimentalBit). ]! !

!MetaStackInterpreter methodsFor: 'meta-jump-accesors' stamp: 'NickPapoulias 5/5/2013 17:09'!
setOverrideBehaviorFor: object

	self inline: true.
	
	objectMemory baseHeader: object put: ((objectMemory baseHeader: object) bitOr: (ExperimentalObjectBit))! !

!MetaStackInterpreter methodsFor: 'meta-jump-accesors' stamp: 'NickPapoulias 5/5/2013 17:09'!
shouldOverrideBehaviorFor: object
	
	self inline: true.
	
	^ (objectMemory isIntegerObject: object) not  & (objectMemory getExperimentalBitOf: object)! !

!MetaStackInterpreter methodsFor: 'meta-protocol-overrides' stamp: 'NickPapoulias 4/30/2013 18:15'!
storeAndPopReceiverVariableBytecode

	| rcvr top |

	self inline: true.

	rcvr := self receiver.

	(self shouldOverrideBehaviorFor: rcvr) ifTrue: [
		top := self internalStackTop.
		self sendMetaWriteAtPutFor: rcvr withIndex: (self integerObjectOf: (currentBytecode bitAnd: 7) + 1) andValue: top.  
	] ifFalse: [
		super storeAndPopReceiverVariableBytecode.
	].

	
! !

!MetaStackInterpreter methodsFor: 'meta-jump-accesors' stamp: 'NickPapoulias 5/5/2013 17:09'!
unsetOverrideBehaviorFor: object
	
	self inline: true.
	
	objectMemory baseHeader: object put: ((objectMemory baseHeader: object) bitAnd: (AllButExperimentalBit)).! !

StackUnixDebugConfig subclass: #MetaStackUnixDebugConfig
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'UnixMetaStack'!

!MetaStackUnixDebugConfig methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/6/2013 16:13'!
interpreterClass
	^ MetaStackInterpreter! !

StackUnixConfig subclass: #MetaStackUnixConfig
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'UnixMetaStack'!

!MetaStackUnixConfig methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/6/2013 16:13'!
interpreterClass
	^ MetaStackInterpreter! !
MetaStackInterpreter initialize!
